unit DEMForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DSpGlobals, DSpViewer, StdCtrls, ExtCtrls, ComCtrls, Menus, DSpPFA,
  ExtDlgs, DSpDataMgr;

type
  TArraySingle = array [0..MaxInt div 4 - 1] of single;
  PArraySingle = ^TArraySingle;

  TForm1 = class(TForm)
    StatusBar: TStatusBar;
    ViewerMenu: TMainMenu;
    mFile: TMenuItem;
    mHelp: TMenuItem;
    mHelpHelp: TMenuItem;
    N1: TMenuItem;
    mHelpAbout: TMenuItem;
    mFileOpen: TMenuItem;
    mFileSave: TMenuItem;
    mFileClose: TMenuItem;
    N2: TMenuItem;
    mFileExit: TMenuItem;
    mView: TMenuItem;
    mViewZoom: TMenuItem;
    mViewZoomAll: TMenuItem;
    mViewZoomSelected: TMenuItem;
    N4: TMenuItem;
    mViewZoomFull: TMenuItem;
    mViewZoomBox: TMenuItem;
    mViewToolbars: TMenuItem;
    mViewToolbarsStatusbar: TMenuItem;
    N5: TMenuItem;
    mViewDataManager: TMenuItem;
    mViewRefresh: TMenuItem;
    N3: TMenuItem;
    mFileSaveImage: TMenuItem;
    N6: TMenuItem;
    SavePictureDialog1: TSavePictureDialog;
    mViewZoomPan: TMenuItem;
    viewer: TDSpViewer;
    mHelpWeb: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure mFileExitClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mViewZoomAllClick(Sender: TObject);
    procedure mViewZoomFullClick(Sender: TObject);
    procedure mViewZoomBoxClick(Sender: TObject);
    procedure mViewToolbarsStatusbarClick(Sender: TObject);
    procedure mViewDataManagerClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure mViewRefreshClick(Sender: TObject);
    procedure mHelpHelpClick(Sender: TObject);
    procedure mHelpAboutClick(Sender: TObject);
    procedure mFileSaveImageClick(Sender: TObject);
    procedure mViewZoomPanClick(Sender: TObject);
    procedure DisplayNewScale(Sender: TObject; p: double);
    procedure GetCoordinates(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure mHelpWebClick(Sender: TObject);
  private
    FDataMgr: TDSpDataMgr;
    FFullScreen  : boolean;            // flag to indicate fullscreen mode
  public
    procedure SaveAsImage(bmp: TBitmap; const fn: string);
  end;

var
  Form1: TForm1;

implementation
uses ShellAPI, Math, DSpUtils, DataManager, DSpAboutForm, JPEG;
{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
  FDataMgr := TDSpDataMgr.Create;
end;

procedure TForm1.GetCoordinates(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  StatusBar.Panels[0].Text := Format('R: %.4d C: %.4d', [Y, X]);
  StatusBar.Panels[1].Text := Format('Y: %6.2f X: %6.2f', [viewer.ProjectedY[Y], viewer.ProjectedX[X]{, single(pointer(integer(FData) + (Y * FColumns + X) * 4)^)}]);
end;

procedure TForm1.DisplayNewScale(Sender: TObject; p: double);
begin
  StatusBar.Panels[2].Text := Format('Scale 1:%3.f', [p]);
end;

procedure TForm1.mFileExitClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FDataMgr.Free;
end;

procedure TForm1.mViewZoomAllClick(Sender: TObject);
begin
  viewer.ZoomToAll;
end;

procedure TForm1.mViewZoomFullClick(Sender: TObject);
// Toggles full screen mode
begin
  FFullScreen := not FFullScreen;
  if FFullScreen then begin
    SetWindowLong(Handle, GWL_STYLE, GetWindowLong(Handle, GWL_STYLE) and not WS_CAPTION);
    WindowState := wsMaximized;
  end else begin
    SetWindowLong(Handle, GWL_STYLE, GetWindowLong(Handle, GWL_STYLE) or WS_CAPTION);
    WindowState := wsNormal;
  end;
  Refresh;
end;

procedure TForm1.mViewZoomBoxClick(Sender: TObject);
begin
  viewer.StartZoom;
end;

procedure TForm1.mViewZoomPanClick(Sender: TObject);
begin
  viewer.StartPan;
end;

procedure TForm1.mViewToolbarsStatusbarClick(Sender: TObject);
begin
  StatusBar.Visible := not StatusBar.Visible;
  mViewToolbarsStatusbar.Checked := not mViewToolbarsStatusbar.Checked;
end;

procedure TForm1.mViewDataManagerClick(Sender: TObject);
begin
  DataMgmtForm.DataMgr := FDataMgr;
  DataMgmtForm.Viewer := viewer;
  DataMgmtForm.Show;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  mViewZoom.Enabled := FDataMgr.DataSets > 0;
end;

procedure TForm1.mViewRefreshClick(Sender: TObject);
begin
  viewer.Refresh;
end;

procedure TForm1.mHelpHelpClick(Sender: TObject);
begin
  ShellExecute(0, 'open', 'http://dspatial.sourceforge.net/help/demo022.html', nil, '.', SW_SHOWDEFAULT);
end;

procedure TForm1.mHelpAboutClick(Sender: TObject);
begin
  AboutBox := TDSpAboutBox.Create(self);
  AboutBox.ShowModal;
  AboutBox.Free;
end;

procedure TForm1.mFileSaveImageClick(Sender: TObject);
var bmp: TBitmap;
    r: TRect;
begin
  if SavePictureDialog1.Execute then begin
    bmp := TBitmap.Create;
    try
      bmp.PixelFormat := pf24bit;
      bmp.Width := viewer.Width;
      bmp.Height := viewer.Height;
      r := Rect(0, 0, bmp.Width, bmp.Height);
      bmp.Canvas.CopyRect(r, viewer.Canvas, r);
      SaveAsImage(bmp, SavePictureDialog1.Files[0]);
    finally
      bmp.Free;
    end;
  end;
end;

procedure TForm1.mHelpWebClick(Sender: TObject);
begin
  ShellExecute(0, 'open', 'http://dspatial.sourceforge.net/', nil, '.', SW_SHOWDEFAULT);
end;

procedure TForm1.SaveAsImage(bmp: TBitmap; const fn: string);
var ext: string;
{    png: TPNGObject;}
    jpg: TJPEGImage;
begin
  ext := ExtractFileExt(fn);
  {if ext = '.png' then begin
    png := TPNGObject.Create;
    try
      png.Assign(bmp);
      png.SaveToFile(fn);
    finally
      png.Free;
    end;
  end else} if ext = '.jpg' then begin
    jpg := TJPEGImage.Create;
    try
      jpg.Assign(bmp);
      jpg.CompressionQuality := 80;
      jpg.Compress;
      jpg.SaveToFile(fn);
    finally
      jpg.Free;
    end;
  end else if ext = '.bmp' then bmp.SaveToFile(fn);
end;

end.
