unit DataManager;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ImgList, ComCtrls, ExtCtrls, Menus, DSpGlobals, DSpPFA, DSpViewer,
  StdCtrls, DSpCoordinateSystem, Grids, DSpRRT, ExtDlgs, DSpDataMgr, DSpDataSet;

type
  TDataMgmtForm = class(TForm)
    tvDataSets: TTreeView;
    ilDataSets: TImageList;
    StatusBar1: TStatusBar;
    mnuDataMgr: TMainMenu;
    mFile: TMenuItem;
    mFileAddDatasets: TMenuItem;
    mFileCloseDataset: TMenuItem;
    N1: TMenuItem;
    mFileClose: TMenuItem;
    mHelp: TMenuItem;
    Splitter1: TSplitter;
    OpenDialog1: TOpenDialog;
    pmDataSets: TPopupMenu;
    pmDataSetsAdd: TMenuItem;
    N2: TMenuItem;
    pmDataSetsRefresh: TMenuItem;
    pmDataSetsSort: TMenuItem;
    pmDataSetsBar: TMenuItem;
    pmDataSetsClose: TMenuItem;
    pmDataSetsRenderer: TMenuItem;
    pmDataSetsProperties: TMenuItem;
    mView: TMenuItem;
    mViewRefresh: TMenuItem;
    mViewSort: TMenuItem;
    pcInfo: TPageControl;
    tsInfo: TTabSheet;
    tsViewer: TTabSheet;
    mmInfo: TMemo;
    pmRenderer: TPopupMenu;
    pmRendererDelete: TMenuItem;
    N3: TMenuItem;
    pmRendererUp: TMenuItem;
    pmRendererDown: TMenuItem;
    tvRenderers: TTreeView;
    Splitter2: TSplitter;
    Panel1: TPanel;
    gbRenderers: TGroupBox;
    cbRenderers: TComboBox;
    pnlMinMax: TPanel;
    lblRRTmin: TLabel;
    edRRTMin: TEdit;
    edRRTMax: TEdit;
    Label2: TLabel;
    pnlRamp: TPanel;
    cmdRendUpdate: TButton;
    imgBackground: TImage;
    Label3: TLabel;
    ColorDialog1: TColorDialog;
    dgRamp: TDrawGrid;
    cmdLoadPalette: TButton;
    cmdSavePalette: TButton;
    pmRamp: TPopupMenu;
    pmRampAbove: TMenuItem;
    pmRampBelow: TMenuItem;
    pmRampDelete: TMenuItem;
    N4: TMenuItem;
    pmRampInvert: TMenuItem;
    imgRamp: TImage;
    N5: TMenuItem;
    pmRendererSaveImage: TMenuItem;
    SavePictureDialog1: TSavePictureDialog;
    pmDataSetsHistogram: TMenuItem;
    pmDataSetsSaveAs: TMenuItem;
    SaveDialog1: TSaveDialog;
    pmRendererAdd: TMenuItem;
    pmRendererAdd3Band: TMenuItem;
    pnl3Band: TPanel;
    ed3BandMax: TEdit;
    Label1: TLabel;
    ed3BandMin: TEdit;
    Label4: TLabel;
    procedure AddDataSource(Sender: TObject);
    procedure mFileCloseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure tvDataSetsMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure EditTreeViewLabel(Sender: TObject; Node: TTreeNode;
      var AllowEdit: Boolean);
    procedure tvDataSetsEdited(Sender: TObject; Node: TTreeNode;
      var S: String);
    procedure RefreshDataSetTree(Sender: TObject);
    procedure SortTreeView(Sender: TObject);
    procedure AddSourceToViewer(Sender: TObject);
    procedure DeleteRenderer(Sender: TObject);
    procedure pmRendererUpClick(Sender: TObject);
    procedure pmRendererDownClick(Sender: TObject);
    procedure tvRenderersDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure tvRenderersDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure tvRenderersMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure tvRenderersEdited(Sender: TObject; Node: TTreeNode;
      var S: String);
    procedure cbRenderersChange(Sender: TObject);
    procedure edRRTMaxExit(Sender: TObject);
    procedure edRRTMinExit(Sender: TObject);
    procedure cmdRendUpdateClick(Sender: TObject);
    procedure imgBackgroundClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure dgRampMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure dgRampDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure dgRampGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure dgRampSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure dgRampSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure InsertBreak(Sender: TObject);
    procedure pmRampDeleteClick(Sender: TObject);
    procedure pmRampInvertClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure pmRendererSaveImageClick(Sender: TObject);
    procedure pmDataSetsCloseClick(Sender: TObject);
    procedure pmDataSetsHistogramClick(Sender: TObject);
    procedure pmDataSetsSaveAsClick(Sender: TObject);
    procedure pmRendererAddClick(Sender: TObject);
    procedure pmRendererAdd3BandClick(Sender: TObject);
  private
    FDataMgr  : TDSpDataMgr;
    FViewer   : TDSpViewer;
    FCoordSys : TDSpCoordinateSystem;
    FRampBmp  : TBitmap;                    // bitmap to render ramp
  protected
    procedure  AddDataSetNode(const ds: TDSpDataSet);
    procedure  DisplayNodeInfo(node: TTreeNode);
    procedure  ShowRendererPanel(node: TTreeNode);
    procedure  MakeRamp;
    procedure  RefreshRendererTree(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property   DataMgr: TDSpDataMgr write FDataMgr;
    property   Viewer: TDSpViewer write FViewer;
    property   CoordinateSystem: TDSpCoordinateSystem read FCoordSys;
  end;

var
  DataMgmtForm: TDataMgmtForm;

implementation
{$R *.DFM}
uses DSpUtils, DSpRaster, DSpRenderer, DSpRasterSource, DSpModuleBase,
     DSpHistogram, HistoForm, DEMForm;

constructor TDataMgmtForm.Create(AOwner: TComponent);
begin
  inherited;
  FRampBmp := TBitmap.Create;
  FRampBmp.PixelFormat := pf24bit;
end;

destructor TDataMgmtForm.Destroy;
begin
  FRampBmp.Free;
  if Assigned(FCoordSys) then FCoordSys.Release;
  inherited;
end;

procedure TDataMgmtForm.AddDataSource(Sender: TObject);
var h: TDSpSourceHandler;
    i, num: integer;
    ds: TDSpDataSet;
begin
  OpenDialog1.Filter := PFA.FilterString(dspReadRaster);
  if OpenDialog1.Execute then begin
    // Open the data source, there is precisely one. This is a reference to an
    // instance owned by the PFA. Call h.Close when done. The handler
    // persists at the level of the PFA until all data sets h supports are
    // released. Since at the application level the handler only needs to
    // open and delete data sets, the handler should always be closed after
    // the necessary data sets are created from the source.
    h := PFA.OpenSource(OpenDialog1.Files[0]);
    try
      // Hopefully the source has some data...
      num := h.DataSets;
      if num = 0 then Exit;

      // Add data set descriptions to the tvDataSets tree.
      for i := 0 to num - 1 do begin
        ds := FDataMgr.OpenDataSet(h, i);
        AddDataSetNode(ds);
      end;

      tvDataSets.TopItem.Expand(false);
    finally
      h.Close;     // This only decreases the reference count. The source
    end;           // object persists at the PFA until all data sets are closed.
  end;
end;

procedure TDataMgmtForm.mFileCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TDataMgmtForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FViewer.Refresh;
  FViewer := nil;
  FDataMgr := nil;
  Action := caHide;
end;

procedure TDataMgmtForm.FormShow(Sender: TObject);
begin
  Assert(FDataMgr <> nil, 'DataMgmtForm.Show: FDataMgr not initialized');
  RefreshDataSetTree(self);
end;

procedure TDataMgmtForm.tvDataSetsMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var node: TTreeNode;
begin
  node := tvDataSets.GetNodeAt(X, Y);

  if Button = mbRight then begin
    if (node = nil) or (node.Level = 0) then begin
      pmDataSetsBar.Visible := false;
      pmDataSetsClose.Visible := false;
      pmDataSetsSaveAs.Visible := false;
      pmDataSetsRenderer.Visible := false;
      pmDataSetsHistogram.Visible := false;
      pmDataSetsProperties.Visible := false;
    end else begin
      tvDataSets.Selected := node;
      pmDataSetsBar.Visible := true;
      pmDataSetsClose.Visible := node.Level = 1;
      pmDataSetsSaveAs.Visible := node.Level = 1;
      pmDataSetsHistogram.Visible := node.Level = 2;
      pmDataSetsRenderer.Visible := node.Level = 2;
      pmDataSetsProperties.Visible := true;
    end;
    pmDataSets.Popup(ClientOrigin.x+X, clientOrigin.y+Y+10);
  end else if Button = mbLeft then begin
    mmInfo.Clear;
    if (node = nil) or (node.Level = 0) then Exit;
    DisplayNodeInfo(node);
    if (TObject(node.Parent.Data) is TDSpRaster) then
      if not Assigned(FCoordSys) or FCoordSys.IsCompatible(TDSpRaster(node.Parent.Data).CoordinateSystem) then
        tvDataSets.BeginDrag(false);
  end;
end;

procedure TDataMgmtForm.EditTreeViewLabel(Sender: TObject; Node: TTreeNode;
  var AllowEdit: Boolean);
begin
  AllowEdit := Node <> (Sender as TTreeView).TopItem;
end;

procedure TDataMgmtForm.tvDataSetsEdited(Sender: TObject; Node: TTreeNode;
  var S: String);
begin
  if TObject(Node.Data) is TDSpDataSet then TDSpDataSet(Node.Data).Name := S
  else if TObject(Node.Parent.Data) is TDSpRaster then TDSpRaster(Node.Parent.Data).Band[integer(node.Data)].Name := S;
end;

procedure TDataMgmtForm.AddDataSetNode(const ds: TDSpDataSet);
var node, child: TTreeNode;
    b: integer;
begin
  node := tvDataSets.Items.AddChild(tvDataSets.TopItem, ds.Name);

  if ds is TDSpRaster then begin
    node.ImageIndex := 5;
    node.SelectedIndex := 5;
    node.Data := ds;
    if TDSpRaster(ds).Bands > 0 then
      for b := 0 to TDSpRaster(ds).Bands - 1 do begin
        child := tvDataSets.Items.AddChild(node, TDSpRaster(ds).Band[b].Name);
        if TDSpRaster(ds).BandType[b] = dspRasterBandSurface then begin
          child.ImageIndex := 3;
          child.SelectedIndex := 3;
        end else begin
          child.ImageIndex := 2;
          child.SelectedIndex := 2;
       end;
        child.Data := pointer(b);
      end;
  end;
end;

procedure TDataMgmtForm.RefreshDataSetTree(Sender: TObject);
var node: TTreeNode;
    i: integer;
begin
  tvDatasets.Items.Clear;
  node := tvDataSets.Items.Add(nil, 'Data sets');
  node.ImageIndex := 9;
  node.SelectedIndex := 9;
  if FDataMgr.DataSets > 0 then
    for i := 0 to FDataMgr.DataSets - 1 do AddDataSetNode(FDataMgr.DataSet[i]);
  tvDataSets.TopItem.Expand(false);
  FViewer.Refresh;
end;

procedure TDataMgmtForm.RefreshRendererTree(Sender: TObject);
var node, child: TTreeNode;
    ch: array[0..3] of TTreeNode;
    i, b, c: integer;
    band: TDSpRasterBand;
    renderers: TList;
    rend: TDSpRenderer;
begin
  tvRenderers.Items.Clear;
  node := tvRenderers.Items.Add(nil, 'Renderers');
  node.ImageIndex := 2;
  node.SelectedIndex := 2;

  renderers := FViewer.Renderers;
  if renderers.Count = 0 then Exit;
  for i := 0 to renderers.Count - 1 do begin
    rend := TDSpRenderer(renderers[i]);
    node := tvRenderers.Items.AddChildObject(tvRenderers.TopItem, rend.Name, rend);
    node.ImageIndex := 2;
    node.SelectedIndex := 2;
    if (rend is TDSpRasterRendererSimple) and (TDSpRasterRendererSimple(rend).Bands > 0) then
      for b := 0 to TDSpRasterRendererSimple(rend).Bands - 1 do begin
        band := TDSpRasterRendererSimple(rend).Band[b];
        child := tvRenderers.Items.AddChildObject(node, band.Name, band);
        child.ImageIndex := 3;
        child.SelectedIndex := 3;
      end
    else if (rend is TDSpRasterRenderer3Channel) then begin
      ch[0] := tvRenderers.Items.AddChild(node, 'Red channel');
      ch[0].ImageIndex := 10;
      ch[0].SelectedIndex := 10;
      ch[1] := tvRenderers.Items.AddChild(node, 'Green channel');
      ch[1].ImageIndex := 11;
      ch[1].SelectedIndex := 11;
      ch[2] := tvRenderers.Items.AddChild(node, 'Blue channel');
      ch[2].ImageIndex := 12;
      ch[2].SelectedIndex := 12;
      ch[3] := tvRenderers.Items.AddChild(node, 'Pan sharpening');
      ch[3].ImageIndex := 15;
      ch[3].SelectedIndex := 15;
      for c := 0 to 3 do
        if (TDSpRasterRenderer3Channel(rend).Channel[c].Bands > 0) then
          for b := 0 to TDSpRasterRenderer3Channel(rend).Channel[c].Bands - 1 do begin
            band := TDSpRasterRenderer3Channel(rend).Channel[c].Band[b];
            child := tvRenderers.Items.AddChildObject(ch[c], band.Name, band);
            child.ImageIndex := 3;
            child.SelectedIndex := 3;
          end;
    end;
  end;
  tvRenderers.TopItem.Expand(false);
end;

procedure TDataMgmtForm.SortTreeView(Sender: TObject);
begin
  (Sender as TTreeView).SortType := stText;
end;

procedure TDataMgmtForm.AddSourceToViewer(Sender: TObject);
var node, child, grandchild: TTreeNode;
    rend: TDSpRenderer;
begin
  node := tvDataSets.Selected;
  if not Assigned(FCoordSys) then
    if TObject(node.Parent.Data) is TDSpDataSet then
      FCoordSys := TDSpDataSet(node.Parent.Data).CoordinateSystem.AddRef
    else Assert(false, 'must add some code here');

  if TObject(node.Parent.Data) is TDSpRaster then begin
    rend := TDSpRasterRendererSimple.Create(FViewer, node.Parent.Text);
    (rend as TDSpRasterRendererSimple).AddRasterBand(TDSpRaster(node.Parent.Data).Band[integer(node.Data)]);
    FViewer.AddRenderer(rend);
    child := tvRenderers.Items.AddChildObject(tvRenderers.TopItem, node.Parent.Text, rend);
    child.ImageIndex := 2;
    child.SelectedIndex := 2;
    grandchild := tvRenderers.Items.AddChildObject(child, node.Text, TDSpRaster(node.Parent.Data).Band[integer(node.Data)]);
    grandchild.ImageIndex := 3;
    grandchild.SelectedIndex := 3;
    pnlRamp.Visible := false;
    pnlMinMax.Visible := true;
    pnl3Band.Visible := false;
  end;
end;

procedure TDataMgmtForm.DeleteRenderer(Sender: TObject);
var node: TTreeNode;
begin
  node := tvRenderers.Selected;
  case node.Level of
    // Node level 1: raster renderer or feature renderer
    1: if TObject(node.Data) is TDSpRenderer then begin
         TDSpRenderer(node.Data).DissociateFromViewer;
         TDSpRenderer(node.Data).Free;
         if (FViewer.NumRenderers = 0) and Assigned(FCoordSys) then begin
           FCoordSys.Release;
           FCoordSys := nil;
         end;
       end;
    // Node level 2: raster band of single channel renderer
    2: if TObject(node.Parent.Data) is TDSpRasterRendererSimple then
         TDSpRasterRendererSimple(node.Parent.Data).DeleteRasterBand(TDSpRasterBand(node.Data));
    // Node level 3: raster band of 3-channel renderer
    3: if TObject(node.Parent.Parent.Data) is TDSpRasterRenderer3Channel then
         TDSpRasterRenderer3Channel(node.Parent.Parent.Data).DeleteRasterBand(node.Parent.Index, TDSpRasterBand(node.Data));
  end;
  
  tvRenderers.Items.Delete(node);
  ShowRendererPanel(nil);
  FViewer.SetRendererMap;
end;

procedure TDataMgmtForm.pmDataSetsHistogramClick(Sender: TObject);
var node: TTreeNode;
    raster: TDSpRaster;
    band: integer;
    histFrm: THistoFrm;
begin
  node := tvDataSets.Selected;
  if TObject(node.Parent.Data) is TDSpRaster then begin
    raster := TDSpRaster(node.Parent.Data);
    band := integer(node.Data);

    Cursor := crHourglass;
    histFrm := THistoFrm.Create(self);
    histFrm.Caption := Format('Histogram for: %s (band %s)', [raster.Name, raster.Band[band].Name]);
    histFrm.UseRasterBand(raster, band);
    histFrm.Show;
    Cursor := crDefault;
  end;
end;

procedure TDataMgmtForm.pmDataSetsSaveAsClick(Sender: TObject);
var node: TTreeNode;
    h: TDSpSourceHandler;
    dset: TDSpDataSet;
    dsrc, copy: TDSpSourceDataSet;
begin
  node := tvDataSets.Selected;
  dset := TDSpDataSet(node.Data);

  // Get a list of the drivers that will take this data set
  dsrc := dset.Source;
  try
    SaveDialog1.Filter := PFA.SaveAsFilter(dsrc);
  finally
    dsrc.ReadUnlock;     // dsrc is now undefined. It is important to unlock,
  end;                   // because the Save dialog takes ages to execute:
                         // waiting for interactive user input
  SaveDialog1.Title := 'Save the data set into a new source';
  SaveDialog1.FileName := '';
  if SaveDialog1.Execute then begin
    Cursor := crHourglass;

    // Check if an extension is added. Add extension if necessary
    if ExtractFileExt(SaveDialog1.FileName) = '' then begin
      Application.MessageBox('You must add extension manually to the name of the file', 'Add extension', MB_OK);
      Exit;
    end;

    // Create a new source
    h := PFA.CreateSource(SaveDialog1.FileName);

    // Add the data set to the new source
    dsrc := dset.Source;
    try
      copy := h.AddDataSet(dsrc); // This runs in the current thread. May take
    finally                       // a while to complete.
      dsrc.ReadUnlock;  // dsrc is now undefined.
      h.Close;
    end;

    if copy = nil then MessageDlg('Data set could not be saved!', mtError, [mbOK], 0)
    else begin
      // Add the newly created data set to the tree.
      dset := FDataMgr.AddDataSource(copy);
      AddDataSetNode(dset);
    end;
    Cursor := crDefault;
  end;
end;

procedure TDataMgmtForm.pmRendererAddClick(Sender: TObject);
var node: TTreeNode;
    rend: TDSpRasterRendererSimple;
begin
  rend := TDSpRasterRendererSimple.Create(FViewer, 'Raster renderer');
  FViewer.AddRenderer(rend);
  node := tvRenderers.Items.AddChildObject(tvRenderers.TopItem, 'Raster renderer', rend);
  node.ImageIndex := 2;
  node.SelectedIndex := 2;
  tvRenderers.TopItem.Expand(false);

  pnlRamp.Visible := false;
  pnlMinMax.Visible := true;
  pnl3Band.Visible := false;
end;

procedure TDataMgmtForm.pmRendererAdd3BandClick(Sender: TObject);
var node, child: TTreeNode;
    rend: TDSpRasterRenderer3Channel;
begin
  rend := TDSpRasterRenderer3Channel.Create(FViewer, 'Raster renderer');
  FViewer.AddRenderer(rend);
  node := tvRenderers.Items.AddChildObject(tvRenderers.TopItem, 'Raster renderer', rend);
  node.ImageIndex := 2;
  node.SelectedIndex := 2;

  child := tvRenderers.Items.AddChild(node, 'Red channel');
  child.ImageIndex := 10;
  child.SelectedIndex := 10;
  child := tvRenderers.Items.AddChild(node, 'Green channel');
  child.ImageIndex := 11;
  child.SelectedIndex := 11;
  child := tvRenderers.Items.AddChild(node, 'Blue channel');
  child.ImageIndex := 12;
  child.SelectedIndex := 12;
  child := tvRenderers.Items.AddChild(node, 'Pan sharpening');
  child.ImageIndex := 15;
  child.SelectedIndex := 15;
  tvRenderers.TopItem.Expand(false);
  node.Expand(true);

  pnlRamp.Visible := false;
  pnlMinMax.Visible := false;
  pnl3Band.Visible := true;
end;

procedure TDataMgmtForm.pmRendererUpClick(Sender: TObject);
var node, nodeUp: TTreeNode;
    ndx: integer;
begin
  node := tvRenderers.Selected;
  ndx := node.Index;
  nodeUp := node.GetPrevSibling;
  FViewer.SwapRenderers(ndx, ndx-1);
  node.MoveTo(nodeUp, naInsert);
end;

procedure TDataMgmtForm.pmRendererDownClick(Sender: TObject);
var node, nodeDown: TTreeNode;
    ndx: integer;
begin
  node := tvRenderers.Selected;
  ndx := node.Index;
  nodeDown := node.GetNextSibling;
  FViewer.SwapRenderers(ndx, ndx+1);
  nodeDown.MoveTo(node, naInsert);
end;

procedure TDataMgmtForm.tvRenderersDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var node: TTreeNode;
begin
  node := tvRenderers.GetNodeAt(X, Y);
  Accept := false;
  if (Source = tvDataSets) then                                            // so dragged object is renderable
    if (node = nil) or (node = tvRenderers.TopItem) then Accept := true    // not dropped on an existing renderer
    else if (TObject(tvDataSets.Selected.Parent.Data) is TDSpRaster) then  // node must be able to accept raster band
      if (TObject(node.Data) is TDSpRasterRendererSimple) then Accept := true
      else if (node.Level = 2) and (TObject(node.Parent.Data) is TDSpRasterRenderer3Channel) then
        Accept := true;
end;

procedure TDataMgmtForm.tvRenderersDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var node, child: TTreeNode;
begin
  node := tvRenderers.GetNodeAt(X, Y);
  if (node = nil) or (node = tvRenderers.TopItem) then begin
    AddSourceToViewer(Sender); // create a new renderer, update the viewer
    tvRenderers.TopItem.Expand(true);
  end else begin
    if TObject(node.Data) is TDSpRasterRendererSimple then
      TDSpRasterRendererSimple(node.Data).AddRasterBand(TDSpRaster(tvDataSets.Selected.Parent.Data).Band[integer(tvDataSets.Selected.Data)])
    else if TObject(node.Parent.Data) is TDSpRasterRenderer3Channel then begin
      TDSpRasterRenderer3Channel(node.Parent.Data).AddRasterBand(node.Index, TDSpRaster(tvDataSets.Selected.Parent.Data).Band[integer(tvDataSets.Selected.Data)]);
      if TDSpRasterRenderer3Channel(node.Parent.Data).Bands = 1 then begin begin
       if not Assigned(FCoordSys) then
         if TObject(tvDataSets.Selected.Parent.Data) is TDSpDataSet then
           FCoordSys := TDSpDataSet(tvDataSets.Selected.Parent.Data).CoordinateSystem.AddRef
        else Assert(false, 'must add some code here');
        FViewer.SetViewExtent(TDSpRasterRenderer3Channel(node.Parent.Data).DisplayArea);
      end;
      end;
    end;
// ???     else child := tvRenderers.Items.AddChildObject(node, tvDataSets.Selected.Text, tvDataSets.Selected.Data);
    child := tvRenderers.Items.AddChildObject(node, tvDataSets.Selected.Text, TDSpRaster(tvDataSets.Selected.Parent.Data).Band[integer(tvDataSets.Selected.Data)]);
    child.ImageIndex := 3;
    child.SelectedIndex := 3;

  end;
  FViewer.SetRendererMap;
end;

procedure TDataMgmtForm.ShowRendererPanel(node: TTreeNode);
var rend : TDSpRasterRenderer;
begin
  if node = nil then begin
    gbRenderers.Visible := false;
    FViewer.Invalidate;
  end else begin
    tvRenderers.Selected := node;

    if node.Level = 1 then begin
      gbRenderers.Visible := true;
      if TObject(node.Data) is TDSpRasterRendererSimple then begin
        rend := TDSpRasterRendererSimple(node.Data);
        if TDSpRasterRendererSimple(rend).Transformation is TDSpRRTHotToCold then cbRenderers.ItemIndex := 0
        else if TDSpRasterRendererSimple(rend).Transformation is TDSpRRTGrayscale then cbRenderers.ItemIndex := 1
        else begin
          cbRenderers.ItemIndex := 2;
          dgRamp.Tag := longint(TDSpRasterRendererSimple(rend).Transformation);
          dgRamp.RowCount := TDSpRRTSmoothRamp(TDSpRasterRendererSimple(rend).Transformation).Intervals + 1;
          dgRamp.Repaint;
        end;
        pnlMinMax.Visible := cbRenderers.ItemIndex < 2;
        pnlRamp.Visible := not pnlMinMax.Visible;
        pnl3Band.Visible := false;
        edRRTMin.Text := FloatToStr(TDSpRasterRendererSimple(rend).Min);
        edRRTMax.Text := FloatToStr(TDSpRasterRendererSimple(rend).Max);
        MakeRamp;
      end else gbRenderers.Visible := false;
    end else if (node.Level = 2) and (TObject(node.Parent.Data) is TDSpRasterRenderer3Channel) then begin
      gbRenderers.Visible := true;
      pnlMinMax.Visible := false;
      pnlRamp.Visible := false;
      pnl3Band.Visible := true;
      ed3BandMin.Text := FloatToStr(TDSpRasterRenderer3Channel(node.Parent.Data).Min[node.Parent.IndexOf(node)]);
      ed3BandMax.Text := FloatToStr(TDSpRasterRenderer3Channel(node.Parent.Data).Max[node.Parent.IndexOf(node)]);
    end else gbRenderers.Visible := false;
  end;
end;

procedure TDataMgmtForm.tvRenderersMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var node: TTreeNode;
begin
  node := tvRenderers.GetNodeAt(X, Y);
  ShowRendererPanel(node);
  
  if Button = mbRight then begin
    pmRendererAdd.Enabled := (node = nil) or (node = tvRenderers.TopItem);
    pmRendererAdd3Band.Enabled := pmRendererAdd.Enabled;
    pmRendererUp.Enabled := (node <> nil) and (node.Index > 0) and (node.Level = 1);
    pmRendererDown.Enabled := (node <> nil) and (node.Index < node.Parent.Count - 1) and (node.Level = 1);
    pmRendererSaveImage.Enabled := (node <> nil) and (node.Level = 1);

    pmRendererDelete.Enabled := (node <> nil) and (
                                (node.Level = 1) or
                                ((node.Level = 2) and (TObject(node.Parent.Data) is TDSpRasterRendererSimple)) or
                                ((node.Level = 3) and (TObject(node.Parent.Parent.Data) is TDSpRasterRenderer3Channel))
                                );

    pmRenderer.Popup(tvRenderers.ClientOrigin.x+X, tvRenderers.clientOrigin.y+Y+10);
  end;
end;

procedure TDataMgmtForm.tvRenderersEdited(Sender: TObject; Node: TTreeNode;
  var S: String);
begin
  if TObject(Node.Data) is TDSpRenderer then TDSpRenderer(Node.Data).Name := S;
end;

procedure TDataMgmtForm.cbRenderersChange(Sender: TObject);
var rrt: TDSpRRTSingleBand;
begin
  rrt := nil; // satify the compiler
  case cbRenderers.ItemIndex of
    0: rrt := TDSpRRTHotToCold.Create;
    1: rrt := TDSpRRTGrayScale.Create;
    2: rrt := TDSpRRTSmoothRamp.Create;
  end;
  TDSpRasterRendererSimple(tvRenderers.Selected.Data).Transformation := rrt;
  if cbRenderers.ItemIndex = 2 then begin
    pnlRamp.Visible := true;
    dgRamp.Tag := longint(rrt);
    dgRamp.RowCount := TDSpRRTSmoothRamp(rrt).Intervals + 1;
    dgRamp.Repaint;
  end else begin
    pnlRamp.Visible := false;
    cmdRendUpdate.Enabled := true;
  end;
  pnlMinMax.Visible := not pnlRamp.Visible;
  MakeRamp;
end;

procedure TDataMgmtForm.edRRTMaxExit(Sender: TObject);
begin
  try
    StrToFloat(edRRTMax.Text);
  except
    Beep;
    edRRTMax.SetFocus;
  end;
end;

procedure TDataMgmtForm.edRRTMinExit(Sender: TObject);
begin
  try
    StrToFloat(edRRTMin.Text);
  except
    Beep;
    edRRTMin.SetFocus;
  end;
end;

procedure TDataMgmtForm.DisplayNodeInfo(node: TTreeNode);
var str: string;
    ext: TDSpExtent;
    raster: TDSpRaster;
    band: integer;
begin
  case node.Level of
    1: begin
         if TObject(node.Data) is TDSpRaster then begin
           mmInfo.Lines.Add('RASTER INFORMATION');
           mmInfo.Lines.Add('');
           mmInfo.Lines.Add(Format('Raster:'#9#9'%s (ID %d)', [TDSpRaster(node.Data).Name, TDSpRaster(node.Data).ID]));
           mmInfo.Lines.Add(Format('Bands:'#9#9'%d', [TDSpRaster(node.Data).Bands]));
           mmInfo.Lines.Add('');
           mmInfo.Lines.Add(Format('Rows:'#9#9'%d', [TDSpRaster(node.Data).Rows]));
           mmInfo.Lines.Add(Format('Columns:'#9#9'%d', [TDSpRaster(node.Data).Columns]));
         end;

         mmInfo.Lines.Add('');
         mmInfo.Lines.Add(TDSpDataSet(node.Data).DescribeCS);
         mmInfo.Lines.Add('Extent:');
         ext := TDSpDataSet(node.Data).Extent;
         mmInfo.Lines.Add('  North:'#9 + FloatToStrF(ext.North, ffGeneral, 15, 3) +
                          #9#9'East:'#9 + FloatToStrF(ext.East, ffGeneral, 15, 3));
         mmInfo.Lines.Add('  South:'#9 + FloatToStrF(ext.South, ffGeneral, 15, 3) +
                          #9#9'West:'#9 + FloatToStrF(ext.West, ffGeneral, 15, 3));
       end;
    2: begin
         if TObject(node.Parent.Data) is TDSpRaster then begin
           raster := TDSpRaster(node.Parent.Data);
           band := integer(node.Data);
           mmInfo.Lines.Add('RASTER BAND INFORMATION');
           mmInfo.Lines.Add('');
           mmInfo.Lines.Add(Format('Raster:'#9#9'%s (ID %d)', [raster.Name, raster.ID]));
           mmInfo.Lines.Add(Format('Band:'#9#9'%d', [band]));
           mmInfo.Lines.Add('');
           case raster.BandType[band] of
             dspRasterBandSurface : str := 'Band type:'#9'Continuous (surface)';
             dspRasterBandOrdinal : str := 'Band type:'#9'Categorical (classes)';
           end;
           mmInfo.Lines.Add(str);
           case raster.BandDataType[band] of
             dspUInt8    : str := 'Data type:'#9'unsigned 8-bit integer'#13#10#9#9'range 0 ~ 255';
             dspInt8     : str := 'Data type:'#9'signed 8-bit integer'#13#10#9#9'range -128 ~ 127';
             dspUInt16   : str := 'Data type:'#9'unsigned 16-bit integer'#13#10#9#9'range 0 ~ 65,535';
             dspInt16    : str := 'Data type:'#9'signed 16-bit integer'#13#10#9#9'range -32,768 ~ 32,767';
             dspUInt32   : str := 'Data type:'#9'unsigned 32-bit integer'#13#10#9#9'range 0 ~ 4,294,967,295';
             dspInt32    : str := 'Data type:'#9'signed 32-bit integer'#13#10#9#9'range -2,147,483,648 ~ 2,147,483,647';
             dspInt64    : str := 'Data type:'#9'signed 64-bit integer'#13#10#9#9'range -9,223,372,036,854,775,808 ~ 9,223,372,036,854,775,807';
             dspFloat32  : str := 'Data type:'#9'32-bit floating point'#13#10#9#9'7 significant digits'#13#10#9#9'range '#177' 1.5E-45 ~ 3.4E38';
             dspFloat64  : str := 'Data type:'#9'64-bit floating point'#13#10#9#9'15 significant digits'#13#10#9#9'range '#177' 5.0E-324 ~ 1.7E308';
           end;
           mmInfo.Lines.Add(str);
           mmInfo.Lines.Add('');
           mmInfo.Lines.Add(Format('Minimum value:'#9'%.7g', [raster.Band[band].Min]));
           mmInfo.Lines.Add(Format('Maximum value:'#9'%.7g', [raster.Band[band].Max]));
         end;
    end;
  end;
  mmInfo.SelStart := 0;
end;

procedure TDataMgmtForm.cmdRendUpdateClick(Sender: TObject);
var node: TTreeNode;
begin
  node := tvRenderers.Selected;
  if node = nil then Exit;

  case node.Level of
    1: if (TObject(node.Data) is TDSpRasterRendererSimple) and (cbRenderers.ItemIndex < 2) then begin
         TDSpRasterRendererSimple(tvRenderers.Selected.Data).Range(StrToFloat(edRRTMin.Text), StrToFloat(edRRTMax.Text));
         MakeRamp;
       end;
    2: if (TObject(tvRenderers.Selected.Parent.Data) is TDSpRasterRenderer3Channel) then
         TDSpRasterRenderer3Channel(tvRenderers.Selected.Parent.Data).Range(node.Parent.IndexOf(node), StrToFloat(ed3BandMin.Text), StrToFloat(ed3BandMax.Text));
  end;
  FViewer.SetRendererMap;
end;

procedure TDataMgmtForm.imgBackgroundClick(Sender: TObject);
begin
  ColorDialog1.Color := FViewer.Color;
  if ColorDialog1.Execute then begin
    imgBackground.Canvas.Brush.Color := ColorDialog1.Color;
    imgBackground.Canvas.FillRect(imgBackground.ClientRect);
    FViewer.Color := ColorDialog1.Color;
  end;
end;

procedure TDataMgmtForm.FormActivate(Sender: TObject);
begin
  imgBackground.Canvas.Brush.Color := FViewer.Color;
  imgBackground.Canvas.FillRect(imgBackground.ClientRect);
end;

procedure TDataMgmtForm.dgRampMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var xy: TGridCoord;
begin
  xy := dgRamp.MouseCoord(X, Y);
  if xy.Y = 0 then Exit;
  dgRamp.Row := xy.Y;

  if Button = mbLeft then begin
    if xy.X = 0 then begin
      ColorDialog1.Color := TDSpRRTSmoothRamp(dgRamp.Tag).Colour[xy.Y-1];
      if ColorDialog1.Execute then begin
        TDSpRRTSmoothRamp(dgRamp.Tag).Colour[xy.Y-1] := ColorDialog1.Color;
        dgRamp.Repaint;
        MakeRamp;
      end
    end;
  end else begin
    pmRampDelete.Enabled := dgRamp.RowCount > 3;
    pmRamp.Popup(dgRamp.ClientOrigin.X + X, dgRamp.ClientOrigin.Y + Y);
  end;
end;

procedure TDataMgmtForm.dgRampDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
begin
  if ARow = 0 then begin
    if ACol = 0 then dgRamp.Canvas.TextRect(Rect, Rect.Left+2, Rect.Top+2, 'Col')
    else dgRamp.Canvas.TextRect(Rect, Rect.Left+2, Rect.Top+2, 'Value')
  end else begin
    if ACol = 0 then begin
      dgRamp.Canvas.Brush.Color := TDSpRRTSmoothRamp(dgRamp.Tag).Colour[ARow-1];
      dgRamp.Canvas.FillRect(Rect);
    end else begin
      dgRamp.Canvas.TextRect(Rect, rect.Left+2, rect.Top+2, FloatToStrF(TDSpRRTSmoothRamp(dgRamp.Tag).Value[ARow-1], ffGeneral, 10, 3));
    end;
  end;
end;

procedure TDataMgmtForm.dgRampGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: String);
begin
  if (ACol = 1) and (ARow > 0) then
    Value := FloatToStrF(TDSpRRTSmoothRamp(dgRamp.Tag).Value[ARow-1], ffGeneral, 10, 3);
end;

procedure TDataMgmtForm.dgRampSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var ramp: TDSpRRTSmoothRamp;
begin
  Assert(ARow > 0);
  if Value = '' then Exit;
  cmdRendUpdate.Enabled := true;
  try
    ramp := TDSpRRTSmoothRamp(dgRamp.Tag);
    ramp.Value[ARow-1] := StrToFloat(Value);
    if (ARow = 1) or (ARow = dgRamp.RowCount - 1) then
      TDSpRasterRendererSimple(tvRenderers.Selected.Data).Range(ramp.Value[0], ramp.Value[dgRamp.RowCount-2]);
  except
    cmdRendUpdate.Enabled := false;
  end;
end;

procedure TDataMgmtForm.InsertBreak(Sender: TObject);
var ramp: TDSpRRTSmoothRamp;
    ndx: integer;
begin
  ramp := TDSpRRTSmoothRamp(dgRamp.Tag);
  if (Sender as TMenuItem).Name = 'pmRampAbove' then ndx := dgRamp.Row - 1 else ndx := dgRamp.Row;
  ramp.Insert(ndx);
  if (ndx = 0) or (ndx = dgRamp.RowCount - 2) then
    TDSpRasterRendererSimple(tvRenderers.Selected.Data).Range(ramp.Value[0], ramp.Value[dgRamp.RowCount-2]);
  dgRamp.RowCount := ramp.Intervals + 1;
  dgRamp.Repaint;
  MakeRamp;
end;

procedure TDataMgmtForm.dgRampSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect := ACol > 0;
end;

procedure TDataMgmtForm.pmRampDeleteClick(Sender: TObject);
begin
  if TDSpRRTSmoothRamp(dgRamp.Tag).Delete(dgRamp.Row - 1) then begin
    dgRamp.RowCount := TDSpRRTSmoothRamp(dgRamp.Tag).Intervals + 1;
    dgRamp.Repaint;
    MakeRamp;
  end;
end;

procedure TDataMgmtForm.pmRampInvertClick(Sender: TObject);
begin
  TDSpRRTSmoothRamp(dgRamp.Tag).Invert;
  dgRamp.Repaint;
  MakeRamp;
end;

procedure TDataMgmtForm.MakeRamp;
begin
  FRampBmp.Width := imgRamp.Width;
  FRampBmp.Height := imgRamp.Height;
  TDSpRRT(TDSpRasterRendererSimple(tvRenderers.Selected.Data).Transformation).Ramp(FRampBmp, dspVert);
  imgRamp.Picture.Bitmap := FRampBmp;
end;

procedure TDataMgmtForm.FormResize(Sender: TObject);
begin
  if (tvRenderers.Selected <> nil) and (TObject(tvRenderers.Selected.Data) is TDSpRasterRendererSimple)
     and Active then begin
    MakeRamp;
    edRRTmin.Top := pnlMinMax.Height - 24;
    lblRRTmin.Top := pnlMinMax.Height - 21;
  end;
end;

procedure TDataMgmtForm.pmRendererSaveImageClick(Sender: TObject);
var bmp: TBitmap;
begin
  SavePictureDialog1.FileName := '';
  if SavePictureDialog1.Execute then begin
    bmp := TDSpRenderer(tvRenderers.Selected.Data).SaveImage(SavePictureDialog1.Files[0]);
    try
      Form1.SaveAsImage(bmp, SavePictureDialog1.FileName);
    finally
      bmp.Free;
    end;
  end;
end;

procedure TDataMgmtForm.pmDataSetsCloseClick(Sender: TObject);
var node, renderTop, renderNode: TTreeNode;
    ds: TDSpDataSet;
begin
  // Find the data set
  node := tvDataSets.Selected;    // node holds a data set
  Assert(node.Level = 1);
  ds := TDSpDataSet(node.Data);

  // Instruct the renderers to release any of the bands being used
  if (ds is TDSpRaster) and (TDSpRaster(ds).Bands > 0) then begin
    renderTop := tvRenderers.Items[0];
    renderNode := renderTop.GetFirstChild;
    while renderNode <> nil do begin
      if TObject(renderNode.Data) is TDSpRenderer then
        TDSpRenderer(renderNode.Data).CloseDataSet(ds);
      renderNode := renderTop.GetNextChild(renderNode);
    end;
  end;
 
  // Delete the data set from the DataMgr and zap the in-memory object
  FDataMgr.CloseDataSet(ds);

  // Synch the tree views with the current data
  RefreshDataSetTree(tvDataSets);
  RefreshRendererTree(tvRenderers);
  mmInfo.Clear;
end;

end.
