unit HistoForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, DSpHistogram, DSpRaster;

type
  THistoFrm = class(TForm)
    histo: TDSpHistogram;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    FHistoData: TDSpHisto;
  public
    procedure UseRasterBand(const r: TDSpRaster; band: integer);
  end;

var
  HistoFrm: THistoFrm;

implementation
{$R *.dfm}
uses DSpGlobals, DSpModuleBase;

procedure THistoFrm.UseRasterBand(const r: TDSpRaster; band: integer);
var data: pointer;
    min, max: TDSpFloat64;
begin
  min := r.Band[band].Min;
  max := r.Band[band].Max;
  if min > max then begin
    MessageDlg('Cannot produce histogram from this data: Min and Max are unknown.', mtInformation, [mbOk], 0);
    Exit;
  end;

  data := DSpReadRasterBandData(r, band);
  try
    case r.BandDataType[band] of
      dspUInt8  : FHistoData := TDSpHistoUInt8.Create;
      dspInt8   : FHistoData := TDSpHistoInt8.Create;
      dspUInt16 : FHistoData := TDSpHistoUInt16.Create(Round(min), Round(max));
      dspInt16  : FHistoData := TDSpHistoInt16.Create(Round(min), Round(max));
      dspUInt32 : FHistoData := TDSpHistoUInt32.Create(Round(min), Round(max));
      dspInt32  : FHistoData := TDSpHistoInt32.Create(Round(min), Round(max));
      dspInt64  : FHistoData := TDSpHistoInt64.Create(Round(min), Round(max));
      dspFloat32: FHistoData := TDSpHistoFloat32.Create(min, max);
      dspFloat64: FHistoData := TDSpHistoFloat64.Create(min, max);
    end;
    FHistoData.AddData(data, r.Rows * r.Columns);
    histo.MaxValue := FHistoData.DetermineBestMaxValue;
    histo.AssignHisto(FHistoData);
  finally
    FreeMem(data);
  end;
end;

procedure THistoFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Assigned(FHistoData) then FHistoData.Free;
end;

end.
