{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpBlobMath.pas. The Original Code is available through }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ As the most plastic name of this unit implies, this unit contains basic      }
{ operations that operate on blobs of data. Most exposed routines internally   }
{ use functions that operate on specific data types, in order to increase      }
{ speed and to be able to use optimizations using for instance ASM code.       }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 24                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                                                                              }
{ Version       : 0.1  03Nov24 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpBlobMath;
{$I DSpatial.inc}
interface
uses DSpGlobals;

  procedure BlobReplaceValue8Bit(const blob: pointer; num: integer; original, update: pointer);
  procedure BlobReplaceValue16Bit(const blob: pointer; num: integer; original, update: pointer);
  procedure BlobReplaceValue32Bit(const blob: pointer; num: integer; original, update: pointer);
  procedure BlobReplaceValueFloat64(const blob: pointer; num: integer; original, update: TDSpFloat64);

implementation

procedure BlobReplaceValue8Bit(const blob: pointer; num: integer; original, update: pointer);
var i: integer;
    p: PDSpUInt8;
    org, upd: TDSpUInt8;
begin
  org := PDSpUInt8(original)^;
  upd := PDSpUInt8(update)^;
  p := PDSpUInt8(blob);

  for i := 1 to num do begin
    if p^ = org then p^ := upd;
    Inc(p);
  end;
end;

procedure BlobReplaceValue16Bit(const blob: pointer; num: integer; original, update: pointer);
var i: integer;
    p: PDSpUInt16;
    org, upd: TDSpUInt16;
begin
  org := PDSpUInt16(original)^;
  upd := PDSpUInt16(update)^;
  p := PDSpUInt16(blob);

  for i := 1 to num do begin
    if p^ = org then p^ := upd;
    Inc(p);
  end;
end;

procedure BlobReplaceValue32Bit(const blob: pointer; num: integer; original, update: pointer);
var i: integer;
    p: PDSpUInt32;
    org, upd: TDSpUInt32;
begin
  org := PDSpUInt32(original)^;
  upd := PDSpUInt32(update)^;
  p := PDSpUInt32(blob);

  for i := 1 to num do begin
    if p^ = org then p^ := upd;
    Inc(p);
  end;
end;

procedure BlobReplaceValueFloat64(const blob: pointer; num: integer; original, update: TDSpFloat64);
var i: integer;
    p: PDSpFloat64;
begin
  p := PDSpFloat64(blob);
  for i := 1 to num do
    if p^ = original then p^ := update;
end;

end.
