{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRasterBandAnalysis.pas. The Original Code is         }
{ available through the home page of DSpatial at                               }
{ http://dspatial.sourceforge.net.                                             }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit offers basic raster band math. The functions here offered are      }
{ needed by the kernel to characterize raster bands: min value, max value,     }
{ other statistics. Because these functions are typically executed at the level}
{ of the PFA driver, operation is on blobs of data and not through higher level}
{ objects such as TDSpRasterSource.                                            }
{                                                                              }
{ All of the operations in this unit are on a single band, and they use        }
{ chunked data access for effective memory management. Most of the operations  }
{ serialize the instructions to get the most out of the CPU. Several routines  }
{ are actually coded in asm to reduce instruction count and optimize register  }
{ use, particularly the FPU registers. This means that if exceptions are       }
{ thrown by the code, the debugger may not actually find the right spot. Be    }
{ aware that the code assumes that the parameters are valid and consistent     }
{ (i.e. in FastFunction(data: pointer; num: integer) the data parameter had    }
{ better point to num elements or ugly things will happen), so make sure that  }
{ your own code accessing these functions is bug-free.                         }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 December 21                                             }
{                                                                              }
{ Dependencies  : DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.1  03Dec21 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRasterBandAnalysis;
{$I DSpatial.inc}
interface
uses DSpGlobals;

implementation

end.
