{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpWin32.pas. The Original Code is available through    }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ In this unit all routines that are specific to the Win32 OS are implemented. }
{ Use this unit only when you compile DSpatial for the Win32 platform. Use the }
{ DSpLinux unit for compiling Linux executables.                               }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 28                                             }
{                                                                              }
{ Dependencies  : Windows (pervasive)                                          }
{                                                                              }
{ Version       : 0.2  03Nov28 TRect here defined, refers to the Windows unit. }
{                 0.1  03Sep28 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpWin32;
interface
uses Windows;
type
  TRect = Windows.TRect;

  TDSpMutex = class
  private
    FHandle   : THandle;
    FTimeOut  : cardinal;
  public
    constructor Create(wait: integer = 1000);
    destructor  Destroy; override;

    function    Get(wait: integer = 0): boolean;
    function    Release: boolean;
    property    TimeOut: cardinal read FTimeOut write FTimeOut;
  end;

  procedure Yield;

implementation
uses SysUtils;

///////////////////////////////////////////////////////////////////////////////
// TDSpMutex
constructor TDSpMutex.Create(wait: integer = 1000);
begin
  FHandle := CreateMutex(nil, false, '');
  if FHandle = 0 then
    raise Exception.Create('Could not create mutex');

  FTimeOut := wait; // 1 second default time out
end;

destructor TDSpMutex.Destroy;
begin
  Release;
  CloseHandle(FHandle);
end;

function TDSpMutex.Get(wait: integer = 0): boolean;
begin
  if wait > 0 then Result := WaitForSingleObject(FHandle, wait) = WAIT_OBJECT_0
  else Result := WaitForSingleObject(FHandle, FTimeOut) = WAIT_OBJECT_0;
end;

function TDSpMutex.Release: boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

////////////////////////////////////////////////////////////////////////////////
// Miscellaneous routines
procedure Yield;
begin
  Sleep(0);
end;

end.
